module Flag where  

  -- This source code is automatically extracted from the original
  -- .lagda sources. Its not optimized for human consumption, but I've
  -- tried to improve some of the formatting and add some comments
  -- here and there.

  open import Prelude

  {-#  BUILTIN EQUALITY _==_  #-}
  {-#  BUILTIN REFL Refl  #-}

  -- Basic data types
  data Index : Nat -> Set where
    One : forall {n} -> Index (Succ n)
    Next : forall {n} -> Index n -> Index (Succ n)

  data Pebble : Set where
    Red : Pebble
    White : Pebble

  data Buckets : Nat -> Set where
    Nil : Buckets Zero
    Cons : forall {n} -> Pebble -> Buckets n -> Buckets (Succ n)

  -- Monadic functions
  State : Nat -> Set -> Set
  State n a = Buckets n -> Pair a (Buckets n)

  exec : forall {n a} -> State n a -> Buckets n -> Buckets n
  exec f bs = snd (f bs)

  ccase : {a : Set} -> Pebble -> a -> a -> a
  ccase Red r w = r
  ccase White r w = w

  return : forall {n a} -> a -> State n a
  return x bs = (x , bs)

  _>>=_ : forall {n a b} -> State n a -> (a -> State n b) -> State n b
  _>>=_ c f bs = let xbs = c bs in f (fst xbs) (snd xbs)

  _>>_ : forall {n a b} -> State n a -> State n b -> State n b
  _>>_ c1 c2 = c1 >>= \x -> c2

  -- Mini computer operations
  infix 8 _!_
  _!_ : forall {n} -> Buckets n -> Index n -> Pebble
  Nil ! ()
  (Cons p ps) ! One = p
  (Cons p ps) ! (Next i) = ps ! i

  read : forall {n} -> Index n -> State n Pebble
  read i bs = (bs ! i , bs)

  update : forall {n} -> Index n -> Pebble -> Buckets n -> Buckets n
  update One x (Cons p ps) = Cons x ps
  update (Next i) x (Cons p ps) = Cons p (update i x ps)

  write : forall {n} -> Index n -> Pebble -> State n Unit
  write i p bs = (unit , update i p bs)

  swap : forall {n} -> Index n -> Index n -> State n Unit
  swap i j =  read i >>= \pi ->
              read j >>= \pj ->
              write i pj >>
              write j pi

  -- Defining less-than-or-equal on indices
  inj : forall {n} -> Index n -> Index (Succ n)
  inj One = One
  inj (Next i) = Next (inj i)

  data _<=_ : forall {n} -> (i j : Index n) -> Set where
    Base : forall {n} -> {i : Index (Succ n)}->  One <= i
    Step : forall {n} -> {i j : Index n}-> i <= j -> Next i <= Next j

  data Difference : forall {n} -> (i j : Index n) -> Set where
    Same : forall {n} -> (i : Index n) -> Difference i i
    Step : forall {n} -> (i j : Index n) -> 
      Difference i j -> Difference (inj i) (Next j)

  nextDiff : forall {n} (i j : Index n) -> 
    Difference i j -> Difference (Next i) (Next j)
  nextDiff .j .j (Same j) = Same (Next j)
  nextDiff .(inj i) .(Next j) (Step i j p) = 
    Step (Next i) (Next j) (nextDiff i j p) 

  injDiff : forall {n} -> (i j : Index n) -> 
    Difference i j -> Difference (inj i) (inj j)
  injDiff .j .j (Same j) = Same (inj j)
  injDiff .(inj i) .(Next j) (Step i j p) = 
    Step (inj i) (inj j) (injDiff i j p)

  data SortT : forall {n} -> (i j : Index n) -> Set where
    Base :  forall {n} ->  (i : Index n) -> SortT i i 
    Step :  forall {n} ->  (i j : Index n) -> 
      SortT (Next i) (Next j) -> SortT (inj i) (inj j) -> SortT (inj i) (Next j)

  mutual 
    fsSortT : forall {n : Nat} -> (i j : Index n) 
      -> SortT i j -> SortT (Next i) (Next j)
    fsSortT .j .j (Base j) = Base (Next j)
    fsSortT ._ ._ (Step i j p1 p2) = 
      Step (Next i) (Next j) (fsSortT (Next i) (Next j) p1) 
                             (fsSortT (inj i) (inj j) p2)

    injSortT : forall {n : Nat} -> (i j : Index n) 
      -> SortT i j -> SortT (inj i) (inj j)
    injSortT .j .j (Base j) = Base (inj j)
    injSortT .(inj i) .(Next j) (Step i j p1 p2) = 
      Step (inj i) (inj j) (fsSortT (inj i) (inj j) p2) 
                           (injSortT (inj i) (inj j) p2)

  -- Some useful lemmas
  toSortT : {n : Nat} -> (i j : Index n) -> i <= j -> SortT i j
  toSortT (Next i) One ()
  toSortT (Next i) (Next j) (Step p) = fsSortT i j (toSortT i j p) 
  toSortT One j lt = sortTOne j
    where
    sortTOne : forall {n} -> (j : Index (Succ n)) -> SortT One j
    sortTOne {Zero} One = Base One
    sortTOne {Zero} (Next ())
    sortTOne {Succ k} One = Base One
    sortTOne {Succ k} (Next i) =
       Step One i (fsSortT One i (sortTOne i)) (injSortT One i (sortTOne i))

  leqRefl : forall {n} -> (i : Index n) -> i <= i
  leqRefl One = Base
  leqRefl (Next y) = Step (leqRefl y)

  leqSucc : forall {n : Nat} {i j : Index n} -> i <= j -> inj i <= Next j
  leqSucc Base = Base
  leqSucc (Step y) = Step (leqSucc y)

  leqPred : forall {n : Nat} {i j : Index n} -> Next i <= Next j -> i <= j
  leqPred (Step y) = y

  fromSortT : {n : Nat} -> {i j : Index n} -> SortT i j -> i <= j
  fromSortT (Base j) = leqRefl j
  fromSortT (Step i j p1 p2) = leqSucc (leqPred (fromSortT p1))

  _<_ : forall {n} -> Index n -> Index n -> Set
  i < j = (i <= j) /\ (i /= j)

  nextInjective : forall {n} {i j : Index n} -> Next i == Next j -> i == j
  nextInjective Refl = Refl

  ltLeqTrans : forall {n : Nat} {i j k : Index n} -> i < j -> j <= k -> i < k
  ltLeqTrans (Base , neq) Base = magic (neq Refl)
  ltLeqTrans (Base , neq) (Step lt2) = Base , \()
  ltLeqTrans (Step lt1 , neq) (Step lt2) 
    with ltLeqTrans (lt1 , \eq -> neq (cong Next eq)) lt2
  ... | (p1 , p2) = (Step p1 , \eq -> p2 (nextInjective eq))

  leqLtTrans : forall {n : Nat} {i j k : Index n} -> i <= j -> j < k -> i < k
  leqLtTrans Base (Base , neq) = Base , neq
  leqLtTrans Base (Step leq , neq) = Base , \()
  leqLtTrans (Step p1) (Step p2 , neq) 
    with leqLtTrans p1 (p2 , \eq -> neq (cong Next eq))
  ... | (q1 , q2) = (Step q1) , (\eq -> q2 (nextInjective eq))


  -- Sorting
  sort : forall {n} -> (r w : Index n) -> SortT r w -> State n Unit 
  sort .(i) .(i) (Base i) = return unit
  sort .(inj i) .(Next j) (Step i j pDiff pInj)  = 
    read (inj i) >>= \c -> 
    ccase c 
      (sort (Next i) (Next j) pDiff)
      (swap (inj i) (Next j) >>
       sort (inj i) (inj j) pInj)

  polish : (n : Nat) -> State n Unit
  polish Zero = return unit
  polish (Succ k) = sort One (maxIndex k) terminates
    where
      maxIndex : (n : Nat) -> Index (Succ n)
      maxIndex Zero = One
      maxIndex (Succ k) = Next (maxIndex k)
      terminates : SortT One (maxIndex k)
      terminates = toSortT One (maxIndex k) Base

  -- The crucial properties of update, lookup, and swap
  lookupUpdated : forall {n} -> (p : Pebble) (i : Index n) (bs : Buckets n) -> 
    (update i p bs ! i) == p
  lookupUpdated p One (Cons _ xs) = Refl
  lookupUpdated p (Next i) (Cons x xs) = lookupUpdated p i xs

  updatePreservation : forall {n} -> (p : Pebble) (i j : Index n) 
    (bs : Buckets n) -> (j /= i) ->
    (update i p bs ! j) == (bs ! j)
  updatePreservation p One One bs q with q Refl
  ... | () 
  updatePreservation p One (Next j) (Cons x xs) q = Refl
  updatePreservation p (Next i) One (Cons x xs) q = Refl
  updatePreservation p (Next i) (Next j) (Cons x xs) q = 
    updatePreservation p i j xs (\p -> q (cong Next p))

  swapPreservation : forall {n} (i : Index n) (x y : Index n) (bs : Buckets n)
   -> i /= x -> i /= y -> exec (swap x y) bs ! i == bs ! i
  swapPreservation i x y bs neq1 neq2 
    rewrite updatePreservation (bs ! x) y i (update x (bs ! y) bs) neq2
      | updatePreservation (bs ! y) x i bs neq1 = Refl

  -- The invariant our sort function should satisfy
  Invariant : forall {n} -> (r w : Index n) -> Buckets n -> Set
  Invariant r w bs = 
    (forall i -> i < r -> (bs ! i) == Red) 
    /\  (forall i -> w < i -> (bs ! i) == White)
 
  -- Two useful views on indices
  injView : forall {n : Nat} (i : Index (Succ n)) (j : Index n) -> i < Next j 
    -> Either (i == inj j) (i < inj j)
  injView .One One (Base , p2) = Inl Refl
  injView .One (Next y) (Base , p2) =
    Inr (Base , \())
  injView .(Next One) One (Step Base , p2) = magic (p2 Refl)
  injView (Next i) (Next j) (Step lt , neq) 
    with injView i j (lt , \p -> neq (cong Next p))
  ... | Inl q = Inl (cong Next q) 
  ... | Inr (q1 , q2) = Inr (Step q1 , \p -> q2 (nextInjective p))

  fsView : forall {n : Nat} (i : Index n) (j : Index (Succ n)) -> inj i < j 
    -> Either (j == Next i) (Next i < j)
  fsView One One (lt , neq) = magic (neq Refl)
  fsView One (Next One) (lt , p) = Inl Refl
  fsView One (Next (Next y)) (lt , p) = Inr (Step Base , \())
  fsView (Next y) One (() , neq)
  fsView (Next i) (Next j) (Step lt , neq) 
    with fsView i j (lt , \eq -> neq (cong Next eq))
  ... | Inl eq = Inl (cong Next eq)
  ... | Inr (leq' , neq') = Inr (Step leq' , \eq -> neq' (nextInjective eq))

  -- The correctness proof
  sortInv : forall {n} -> (r w : Index n) ( d : SortT r w) ( bs : Buckets n) 
    -> Invariant r w bs
    -> Sigma (Index n) (\m -> Invariant m m (exec (sort r w d) bs) )
  --  Base case
  sortInv .w w (Base .w) bs (redH , whiteH) = Witness w (redH , whiteH)
  --  Split on whether or not bs ! inj i is Red or White
  sortInv .(inj i) .(Next j) (Step i j nextT injT) bs p with inspect (bs ! inj i)
  ... | x with-== eq with (bs ! inj i) | eq
  -- If bs ! inj i was Red, we can quickly re-establish the invariant.
  sortInv .(inj i) .(Next j) (Step i j nextT injT) bs (redH , whiteH)
    | .Red with-== eq | Red | Refl
      = sortInv (Next i) (Next j) nextT bs (lemma , whiteH)
    where
    lemma : forall k -> k < (Next i) -> (bs ! k) == Red
    lemma k lt with k | injView k i lt
    ... | .(inj i) | Inl Refl = eq
    ... | k' | Inr q = redH k' q
  -- If bs ! inj i was White we need to perform a swap
  --   and hence do more work to re-establish our invariant.
  sortInv .(inj i) .(Next j) (Step i j nextT injT) bs (redH , whiteH)
    | .White with-== eq | White | Refl = 
    sortInv (inj i) (inj j) injT bs' (lemmaRed , lemmaWhite)
    where
    bs' = exec (swap (inj i) (Next j)) bs
    neqK : forall k -> k < inj i -> k /= Next j
    neqK k lt = snd (ltLeqTrans lt (fromSortT (Step i j nextT injT)))
    lemmaRed : forall k -> (k < inj i) -> bs' ! k == Red
    lemmaRed k (leq , neq)
      rewrite swapPreservation k (inj i) (Next j) bs neq (neqK k (leq , neq)) 
        = redH k (leq , neq)
    lemmaWhite : forall k -> (inj j < k) -> bs' ! k == White
    lemmaWhite k lt with fsView j k lt
    ... | Inl eqK rewrite eqK | eq
      = lookupUpdated White (Next j) (update (inj i) (bs ! Next j) bs)
    ... | Inr (lt' , neq') 
           rewrite swapPreservation k (inj i) (Next j) bs 
              ((\eq -> snd (leqLtTrans (fromSortT injT) lt) (sym eq))) 
              ((\p -> neq' (sym p))) 
              = whiteH k (lt' , neq')


