module Prelude where

data Nat : Set where
  Zero : Nat
  Succ  : (k : Nat) -> Nat

{-# BUILTIN NATURAL  Nat  #-}
{-# BUILTIN ZERO     Zero #-}
{-# BUILTIN SUC      Succ #-}

data Pair (a b : Set) : Set where
  _,_ : a -> b -> Pair a b

infix 5 _,_
infixr 5 _/\_
_/\_ : Set -> Set -> Set
a /\ b = Pair a b

fst : {A B : Set} -> Pair A B -> A
fst (x , y) = x

snd : {A B : Set} -> Pair A B -> B
snd (x , y) = y

data Unit : Set where
  unit : Unit

data Sigma (A : Set) (B : A -> Set) : Set where
  Witness : (x : A) -> B x -> Sigma A B

data _==_ {a : Set} (x : a) : a -> Set where
  Refl : x == x
infix 6 _==_

data Empty : Set where

magic : forall {a : Set} -> Empty -> a
magic ()

_/=_ : {a : Set} -> a -> a -> Set
x /= y = (x == y) -> Empty

cong : {A B : Set} -> {x y : A} -> (f : A -> B) -> x == y -> f x == f y
cong f Refl = Refl


trans : {A : Set} -> {x y z : A} -> x == y -> y == z -> x == z
trans Refl Refl = Refl

sym : {A : Set} -> {x y : A} -> x == y -> y == x
sym Refl = Refl

data Inspect {a : Set} (x : a) : Set where
  _with-==_ : (y : a) (eq : x == y) -> Inspect x

inspect : forall {a} (x : a) -> Inspect x
inspect x = x with-== Refl

data Either (a b : Set) : Set where
  Inl : a -> Either a b
  Inr : b -> Either a b